/* This file has been prepared for Doxygen automatic documentation generation.*/
/*! \file *********************************************************************
 *
 * \brief
 *      Functions for use of PWM
 *
 *      Contains functions for initializing and controlling PWM output.
 *
 * \par Application note:
 *      AVR458: Charging Li-Ion Batteries with BC100\n
 *      AVR463: Charging NiMH Batteries with BC100
 *
 * \par Documentation
 *      For comprehensive code documentation, supported compilers, compiler 
 *      settings and supported devices see readme.html
 *
 * \author
 *      Atmel Corporation: http://www.atmel.com \n
 *      Support email: avr@atmel.com
 *
 * 
 * $Name$
 * $Revision: 2299 $
 * $RCSfile$
 * $URL: http://svn.norway.atmel.com/AppsAVR8/avr458_Charging_Li-Ion_Batteries_with_BC100/tag/20070904_release_1.0/code/IAR/PWM.c $
 * $Date: 2007-08-23 12:55:51 +0200 (to, 23 aug 2007) $\n
 ******************************************************************************/

#include <avr/io.h>

#include "enums.h"

#include "main.h"
#include "PWM.h"
#include "time.h"


//******************************************************************************
// Functions
//******************************************************************************
/*! \brief Stops PWM output
 *
 */
void PWM_Stop(void)
{
	OCR1B = 0;  // Reset compare level.
	PLLCSR = 0; // Disable PLL, switch to synchronous CLK mode.
	TCCR1A = 0; // Set normal port operation, disable PWM modes.
	TCCR1B = 0; // Stop timer/counter1.
	TCCR1C = 0; // Set normal port operation.
	TCCR1D = 0; // No fault protection, normal waveform.
	OCR1C = 0;  // Reset compare.
	OCR1D = 0;  // Reset compare.
	DT1 = 0;    // No dead time values.
}


/*! \brief Initializes and starts PWM output
 *
 * Initializes timer1 for use as a PWM with a clock rate of 64 MHz.\n
 * Its comparator is connected to PB3 and will output high until timer1 reaches
 * the value of OCR1B. It is then dropped to 0.\n
 * The comparator outputs high again when the counter overflows, which will
 * happen at a rate of 250 kHz.
 */
void PWM_Start(void)
{
	// Clear OC1B on compare match, enable PWM on comparator OCR1B.
	TCCR1A = (1<<COM1B1)|(0<<COM1B0)|(1<<PWM1B);
	
	// Non-inverted PWM, T/C stopped.
	TCCR1B = 0;
	
	// Copy shadow bits, disconnect OC1D.
	TCCR1C = (TCCR1A & 0xF0);
	
	// No fault protection, use phase & frequency correct PWM.
	TCCR1D = (0<<WGM11)|(1<WGM10);
	
	// Does not matter -- PWM6 mode not used.
	TCCR1E = 0;
	
	// Does not matter -- OC1A is disabled.
	OCR1A = 0;
	
	// Set reset compare level. (Offset is used, or JumperCheck() will fail.)
	OCR1B = PWM_OFFSET;
	
	// TOP value for PWM, f(PWM) = 64MHz / 255 = 251kHz.
	OCR1C = 0xFF;
	
	// Does not matter -- OC1D is disabled.
	OCR1D = 0;
	
	// No dead time values.
	DT1 = 0;
	
	// Set PWM port pin to output.
	DDRB |= (1<<PB3);
	
	// Enable PLL, use full speed mode.
	PLLCSR = (0<<LSM) | (1<<PLLE);
	
	// Use general timer and wait 1 ms for PLL lock to settle.
	Time_Set(TIMER_GEN,0,0,1);
	do{ 
	}while(Time_Left(TIMER_GEN));
	
	// Now wait for PLL to lock.
	do{ 
	}while((PLLCSR & (1<<PLOCK)) == 0);

	// Use PLL as clock source.
	PLLCSR |= (1<<PCKE);
	
	// CLK PCK = 64MHz / 1 = 64MHz.
	TCCR1B |= (0<<CS13)|(0<<CS12)|(0<<CS11)|(1<<CS10);
}


/*! \brief Increments the PWM duty cycle, if not already at max
 *
 * \retval TRUE Success, duty cycle could be incremented.
 * \retval FALSE Failure, duty cycle already at maximum.
 */
unsigned char PWM_IncrementDutyCycle(void){

	if (OCR1B < PWM_MAX) {
		OCR1B += 1;
		return(TRUE);
	} else {
		return(FALSE);
	}
}


/*! \brief Decrements the PWM duty cycle, if not already at zero.
 *
 * \retval TRUE Success, duty cycle could be decremented.
 * \retval FALSE Failure, duty cycle already at zero.
 */
unsigned char PWM_DecrementDutyCycle(void){
	
	if (OCR1B > 0)	{
		OCR1B -= 1;
		return(TRUE);
	} else {
		return(FALSE);
	}
}
