/* This file has been prepared for Doxygen automatic documentation generation.*/
/*! \file *********************************************************************
 *
 * \brief
 *      Functions for use of ADC
 *
 *      Contains high level functions for initializing the ADC, interrupt
 *      handling, and treatment of samples.\n
 *      The ADC is set to free running mode and uses an external reference
 *      voltage.\n
 *      To make all sampling take at least 25 clock cycles the ADC is stopped
 *      and restarted by the ISR.
 *
 * \par Application note:
 *      AVR458: Charging Li-Ion Batteries with BC100 \n
 *      AVR463: Charging NiMH Batteries with BC100
 *
 * \par Documentation:
 *      For comprehensive code documentation, supported compilers, compiler
 *      settings and supported devices see readme.html
 *
 * \author
 *      Atmel Corporation: http://www.atmel.com \n
 *      Support email: avr@atmel.com \n
 *      Original author: \n
 *
 * $Name$
 * $Revision: 2299 $
 * $RCSfile$
 * $URL: http://svn.norway.atmel.com/AppsAVR8/avr458_Charging_Li-Ion_Batteries_with_BC100/tag/20070904_release_1.0/code/IAR/ADC.c $
 * $Date: 2007-08-23 12:55:51 +0200 (to, 23 aug 2007) $\n
 ******************************************************************************/

#include <ioavr.h>
#include <inavr.h>

#include "structs.h"

#include "main.h"
#include "ADC.h"


//******************************************************************************
// Variables
//******************************************************************************
// ADC status struct.
//! \brief Holds sampled data and ADC-status
ADC_Status_t ADCS;


/*! \brief Indicates maximum battery voltage.
 *
 * This variable is stored in EEPROM and indicates how much the battery voltage
 * is downscaled by HW before it is sampled. The amount of downscaling depends
 * on the maximum battery voltage, and is necessary to avoid saturation of the
 * ADC (reference voltage is 2.5 V).
 *
 * \note Used by the ADC ISR when calling ScaleU() and ScaleI().
 *
 * \note Defaults to 1, which means 10 V max battery voltage.
 *
 * \note Table of settings:
 * <pre>
 * VBAT_RANGE | Max battery voltage | Jumper setting
 *         0  |             5V      |        1/2
 *         1  |            10V      |        1/4
 *         2  |            20V      |        1/8
 *         3  |            30V      |       1/12
 *         4  |            40V      |       1/16
 * </pre>
 */
// Maximum battery voltage (affects scaling of samples).
__eeprom unsigned char VBAT_RANGE = 1;


//******************************************************************************
// Functions
//******************************************************************************
/*! \brief Interrupt Service routine for ADC.
 *
 * This ISR stores the sampled values in the ADC status-struct, then
 * updates the ADC MUX to the next channel in the scanning-sequence.\n
 * Once the sequence is completed, ADCS.Flag is set and unless
 * ADCS.Halt has been set, the sequence starts over. Otherwise, the ADC 
 * is disabled.\n
 * If the mains voltage is below minimum, ADCS.Mains gets set to FALSE.
 *
 * \note Table of scanning sequence:
 * <pre>
 * Seq |    MUX |  pos I/P |  neg I/P | gain | measure | signed
 * ----+--------+----------+----------+------+---------+-------
 *  01 | 000001 | ADC1/PA1 |      n/a |   1x |     NTC |     no
 *  02 | 000010 | ADC2/PA2 |      n/a |   1x |     RID |     no
 *  03 | 000011 | ADC3/PA4 |      n/a |   1x |    VIN- |     no
 *  04 | 000100 | ADC4/PA5 |      n/a |   1x |    VIN+ |     no
 *  05 | 000101 | ADC5/PA6 |      n/a |   1x |   VBAT- |     no
 *  06 | 000110 | ADC6/PA7 |      n/a |   1x |   VBAT+ |     no
 *  07 | 010010 | ADC4/PA5 | ADC3/PA4 |  20x |     IIN |     no
 *  08 | 010111 | ADC6/PA7 | ADC5/PA6 |  20x |    IBAT |    yes
 * </pre>
 *
 * \todo IIN (#7 in sequence) is never used.
 *
 * \todo Signed is never set. Signed measurements of IBAT will halve the
 * measuring sensitivity, and is therefore not favourable. At the moment,
 * great currents (f.ex. if something happens with the battery) will be
 * interpreted as negative, which might cause unfavourable behaviour during
 * charging (depending on what PWM behaviour is defined), f.ex.
 * ConstantCurrent() will keep increasing the PWM output. This results in an
 * PWM controller error being flagged and the program going into
 * error-state and eventually reinitializing.
 */
#pragma vector=ADC_vect
__interrupt void ADC_ISR(void)
{
	static unsigned char avgIndex = 0;
	unsigned char i, Next, Signed;
	signed int  temp = 0;
	
	Signed = FALSE;  // Presume next conversion is unipolar.
	ADCSRA &= ~(1<<ADEN);  // Stop conversion before handling. This makes all
	  // conversions take at least 25 ADCCLK. (It is restarted later)
	
	// Handle the conversion, depending on what channel it is from, then
	// switch to the next channel in the sequence.
	switch (ADCS.MUX){
		// MUX = 0b000001 => ADC1 (PA1) = NTC
		case 0x01:
			ADCS.rawNTC = ADC;
			Next=0x02;
		break;

		
		// MUX = 0b000010 => ADC2 (PA2) = RID
		case 0x02:
			ADCS.rawRID = ADC;
			Next=0x03;
		break;

		
		// MUX = 0b000011 => ADC3 (PA4) = VIN-
		case 0x03:
			// Supply voltage is always divided by 16.
			ADCS.VIN = ScaleU(4, (unsigned int)ADC);  // Cast because ADC is short.
			
			// Is mains failing?
			if (ADCS.VIN < VIN_MIN) {
				ADCS.Mains = FALSE;
			} else {
				ADCS.Mains = TRUE;
			}
			
			Next=0x05;
		break;

		
		// MUX = 0b000101 => ADC5 (PA6) = VBAT-
		case 0x05:
			ADCS.rawVBAT = ADC;
			
			// Scale voltage according to jumper setting.
			ADCS.VBAT = ScaleU(VBAT_RANGE, (unsigned int)ADC); // ADC is a short.
			Next=0x17;
//			Signed = TRUE;  // Next conversion is bipolar. Halves sensitivity!
		break;


		case 0x17:  // MUX = 0b010111 => 20 x [ADC6(PA7) - ADC5(PA6)] = IBAT
			// If bipolar, from -512 to 0, to 511:
			// 0x200 ... 0x3ff, 0x000, 0x001 ... 0x1FF
		
			// Scale sample according to jumper setting, handle negative numbers.
			if (ADC > 511) {
				ADCS.IBAT = -(signed int)ScaleI(VBAT_RANGE,
				             (1024 - (ADC-ADCS.ADC5_G20_OS)));
			} else if (ADC > 0) {
				ADCS.IBAT = ScaleI(VBAT_RANGE, (ADC-ADCS.ADC5_G20_OS));
			} else {
				ADCS.IBAT = 0;
			}

			// Insert sample of battery current into the averaging-array
			// (overwriting the oldest sample), then recalculate and store the
			// average. This is the last conversion in the sequence, so
			// flag a complete ADC-cycle and restart sequence.
			ADCS.discIBAT[(avgIndex++ & 0x03)] = ADCS.IBAT;
			for (i = 0; i < 4 ; i++) {
				temp += ADCS.discIBAT[i];
			}
			
			ADCS.avgIBAT = (temp / 4);
			
			ADCS.Flag = TRUE;
			Next=0x01;
			Signed = FALSE;  // This is the only bipolar conversion.
		break;

		
		default:  // Should not happen. (Invalid MUX-channel)
			Next=0x01;  // Start at the beginning of sequence.
		break;
	}
	
	// Update MUX to next channel in sequence, set a bipolar conversion if
	// this has been flagged.
	ADCS.MUX = Next;                    
	ADMUX = (1<<REFS0) + ADCS.MUX;      

	if (Signed)	{
	  ADCSRB |= (1<<BIN);               
	} else {
	  ADCSRB &= ~(1<<BIN);              
	}

	// Re-enable the ADC unless a halt has been flagged and a conversion
	// cycle has completed.
	if (!((ADCS.Halt) && (ADCS.Flag))) {
	  ADCSRA |= (1<<ADEN)|(1<<ADSC);    
	}
}


/*! \brief Scales sample to represent "actual voltage" in mV.
 *
 * This function returns the actual sampled voltage, scaled according
 * to the jumper settings.
 *
 * \param setting Indicates what downscaling was used.
 * \param data The sampled value.
 *
 * \note Table for setting-parameter:\n
 * <pre>
 * Presume VREF = 2.5V and Gain = 1x.
 * => Resolution @ 1/1 = 2.5V / 1024 = 2.4414 mV/LSB
 * setting | source |   R1 | R2/(R1+R2) | UADC(LSB) | U(MAX)
 * --------+--------+------+------------+-----------+-------
 *     N/A |        |    - |       -    |   2.441mV |  2.50V
 *       0 |   VBAT |  10k |     1/2    |   4.883mV |  5.00V
 *       1 |   VBAT |  30k |     1/4    |   9.766mV |  9.99V
 *       2 |   VBAT |  70k |     1/8    |   19.53mV | 19.98V
 *       3 |   VBAT | 110k |    1/12    |   29.30mV | 29.97V
 *       4 |   VBAT | 150k |    1/16    |   39.06mV | 39.96V
 *       4 |    VIN | 150k |    1/16    |   39.06mV | 39.96V
 * </pre>
 */
unsigned int ScaleU(unsigned char setting, unsigned int data)
{
	// Temporary variable needed.
	unsigned int scaled = 0;

	// Jumper setting 3: mV/LSB = 29.30 ~= 29 + 1/4 + 1/16
	if (setting == 3)	{
		scaled = 29 * data;
		scaled += (data >> 2);
		scaled += (data >> 4);
	} else {
		// Jumper setting 4: mV/LSB = 39.06 ~= 39 + 1/16
		scaled = 39 * data;
		scaled += (data >> 4);
		
		if (setting <3) {
			// Jumper setting 0: mV/LSB = 4.883 = 39.06 / 8
			//                1: mV/LSB = 9.766 = 39.06 / 4
			//                2: mV/LSB = 19.53 = 39.06 / 2
			scaled = (scaled >> (3-setting));
		}
	}

	return(scaled);
}


/*! \brief Scales sample to represent "actual current" in mA.
 *
 * This function returns the actual sampled current, scaled according
 * to the jumper settings.
 *
 * \param setting Indicates what downscaling was used.
 * \param data The sampled value.
 *
 * \note Table for setting-parameter:\n
 * <pre>
 * Presume VREF = 2.5V and Gain = 1x or 20x.
 * => Resolution(U) @ (1/1 and 20x) = 2.5V / (GAIN x 1024) = 0.1221 mV/LSB
 * => Resolution(I) = Resolution(U) / Rshunt = Resolution(U) / 0.07
 * Setting |   R1 | R2/(R1+R2) |   U(LSB) |   I(LSB) | I(MAX) | Gain
 * --------+------+------------+----------+----------+--------+-----
 *     N/A |    - |       -    | 0.1221mV |  1.744mA |  1.78A |  20x
 *       0 |  10k |     1/2    | 0.2442mV |  3.489mA |  3.57A |  20x
 *       1 |  30k |     1/4    | 0.4884mV |  6.978mA |  7.14A |  20x
 *       2 |  70k |     1/8    | 0.9768mV | 13.955mA |  14.3A |  20x
 *       3 | 110k |    1/12    | 1.4652mV | 20.931mA |  21.4A |  20x
 *       4 | 150k |    1/16    | 1.9536mV | 27.909mA |  28.5A |  20x
 *       5 |  10k |     1/2    | 2.4414mV | 34.877mA |  35.7A |   1x
 * </pre>
 */
unsigned int ScaleI(unsigned char setting, unsigned int data)
{
	// Temporary variable needed.
	unsigned int  scaled = 0;
	
	// Jumper setting 3: mA/LSB = 20.931mA ~= 21 - 1/16 + 1/128
	if (setting == 3) {
		scaled = 21 * data;
		scaled -= (data >> 4);
		scaled += (data >> 7);
	}	else	{ // Jumper setting 4: mA/LSB = 27.909mA ~= 28 - 1/8 + 1/32
		scaled = 28 * data;
		scaled -= (data >> 3);
		scaled += (data >> 5);
		
		if (setting <3) {
			// Jumper setting 0: mA/LSB = 3.489mA = 27.909 / 8
			//                1: mA/LSB = 6.978mA = 27.909 / 4
			//                2: mA/LSB = 13.955mA = 27.909 / 2
			scaled = (scaled >> (3-setting));
		}
	}
	
	return(scaled);
}


/*! \brief Waits for two full cycles of ADC-conversions to occur.
 *
 * This function clears the cycle complete-flag, then waits for it to be set
 * again. This is then repeated once before the function exits.
 * 
 */
void ADC_Wait(void)
{
	// Clear ADC flag and wait for cycle to complete.
	ADCS.Flag = FALSE;              
	do {
	} while (ADCS.Flag == FALSE);      
	
	// Repeat, so we are sure the data beong to the same cycle.
	ADCS.Flag = FALSE;              
	do {
	} while (ADCS.Flag == FALSE);      
}


/*! \brief Initializes ADC and input pins.
 *
 * This function initializes the ADC to free running mode, sampling from
 * PA1/2/4/5/6/7, and using an external reference voltage (PA3).\n
 * It also measures and stores calibration data for offset.
 *
 * \todo Odd offset measurement for ADC3_G20_OS? It is never used anyway.
 *
 * \note Table of MUX settings for offset measurement:
 * <pre>
 *    Ch | Pin |    Gain |    MUX
 * ------+-----+---------+-------
 *  ADC1 | PA1 |     20x | 001101
 *  ADC3 | PA4 |     20x | 010001
 *  ADC5 | PA6 |     20x | 010110
 *  ADC9 | PB6 |     20x | 011011
 *  ADC0 | PA0 | 20x/32x | 111000
 *  ADC0 | PA0 |   1x/8x | 111001
 *  ADC1 | PA1 | 20x/32x | 111010
 *  ADC2 | PA2 | 20x/32x | 111011
 *  ADC4 | PA5 | 20x/32x | 111100
 *  ADC5 | PA6 | 20x/32x | 111101
 *  ADC6 | PA7 | 20x/32x | 111110
 * </pre>
 */
void ADC_Init(void)
{
	unsigned char i;

	__disable_interrupt();

	ADCS.Halt = FALSE; // Enable consecutive runs of ADC.

	// Configure ADC pins (inputs and disabled pull-ups).
	DDRA &= ~((1<<PA1)|(1<<PA2)|(1<<PA4)|(1<<PA5)|(1<<PA6)|(1<<PA7));
	PORTA &= ~((1<<PA1)|(1<<PA2)|(1<<PA4)|(1<<PA5)|(1<<PA6)|(1<<PA7));

	// Set ADC3 as reference, and MUX to measure the same pin.
	ADMUX = (1<<REFS0) | (1<<MUX0) | (1<<MUX1);
	
	ADCSRB = 0;

	// Start conversion, no interrupt (disable ADC-ISR).
	ADCSRA = (1<<ADEN) | (1<<ADSC) | ADC_PRESCALER; 

	do { // Wait for conversion to finish.
	} while (!(ADCSRA & (1<<ADIF)));

	ADCSRA |= (1<<ADIF);  // Clear ADC interrupt flag manually.

	ADCS.ADC3_G20_OS = ADC;  // Save the sampled offset.

	ADMUX = (1<<REFS0) | 0x16;  // ADC5/ADC5 (external ref.), 20x
	
	// Start conversion, no interrupt. ADC_PRESCALER is defined in ADC.h.
	ADCSRA = (1<<ADEN) | (1<<ADSC) | ADC_PRESCALER; 

	do { // Wait for conversion to finish.
	} while (!(ADCSRA & (1<<ADIF)));

	ADCSRA |= (1<<ADIF);  // Clear ADC interrupt flag.

	ADCS.ADC5_G20_OS = ADC;  // Save the sampled offset.

	// Reset the ADC-cycle.
	ADCS.Flag = FALSE;	
	ADCS.MUX = 0x01;                    
	ADMUX = (1<<REFS0) | ADCS.MUX;      

	// Clear averaged battery current and the discrete readings.
	ADCS.avgIBAT = 0;
	
	for (i = 0; i < 4; i++) {
		ADCS.discIBAT[i] = 0;             
	}
	
	// Re-enable the ADC and ISR.
	ADCSRA=(1<<ADEN)|(1<<ADSC)|(1<<ADIE)|ADC_PRESCALER;
	
	__enable_interrupt();

	// Get a complete cycle of data before returning.
	ADC_Wait();
}
